@@ -94,7 +94,7 @@ const char *get_as_path (const char *query) {
 
 		while (isspace (*p))  p++;
 		as = p;
-		while (!isspace (*p))  p++;
+		while (*p && !isspace (*p))  p++;
 		*p = '\0';
 
 		if (prefix > best_prefix) {
