@@ -62,14 +62,18 @@
 #define DEF_NUM_PROBES	3
 #define DEF_WAIT_SECS	5.0
 #define DEF_SEND_SECS	0
-#define DEF_PACKET_LEN	40
+#define DEF_DATA_LEN	40	/*  all but IP header...  */
 #define MAX_PACKET_LEN	65000
+#ifndef DEF_AF
 #define DEF_AF		AF_INET
+#endif
+
+#define ttl2hops(X)	(((X) <= 64 ? 65 : ((X) <= 128 ? 129 : 256)) - (X))
 
 
 static char version_string[] = "Modern traceroute for Linux, "
 				"version " _TEXT(VERSION) ", " __DATE__
-				"\nCopyright (c) 2006  Dmitry Butskoy, "
+				"\nCopyright (c) 2008  Dmitry Butskoy, "
 				"  License: GPL v2 or any later";
 static int debug = 0;
 static unsigned int first_hop = 1;
@@ -82,16 +86,22 @@ static int num_gateways = 0;
 static unsigned char *rtbuf = NULL;
 static size_t rtbuf_len = 0;
 
+static size_t header_len = 0;
+static size_t data_len = 0;
+
 static int dontfrag = 0;
 static int noresolve = 0;
+static int extension = 0;
 static int as_lookups = 0;
 static unsigned int dst_port_seq = 0;
 static unsigned int tos = 0;
 static unsigned int flow_label = 0;
 static int noroute = 0;
-static unsigned int packet_len = DEF_PACKET_LEN;
+static int packet_len = -1;
 static double wait_secs = DEF_WAIT_SECS;
 static double send_secs = DEF_SEND_SECS;
+static int mtudisc = 0;
+static int backward = 0;
 
 static sockaddr_any dst_addr = {{ 0, }, };
 static char *dst_name = NULL;
@@ -414,7 +424,7 @@ static CLIF_option option_list[] = {
 	{ "6", 0, 0, "Use IPv6", set_af, (void *) 6, 0, 0 },
 	{ "d", "debug", 0, "Enable socket level debugging",
 			CLIF_set_flag, &debug, 0, 0 },
-	{ "F", "dont-fragment", 0, "Set DF (don't fragment bit) on",
+	{ "F", "dont-fragment", 0, "Do not fragment packets",
 			CLIF_set_flag, &dontfrag, 0, CLIF_ABBREV },
 	{ "f", "first", "first_ttl", "Start from the %s hop (instead from 1)",
 			CLIF_set_uint, &first_hop, 0, 0 },
@@ -472,6 +482,9 @@ static CLIF_option option_list[] = {
 			    "in milliseconds, else it is a number of seconds "
 			    "(float point values allowed too)",
 			    CLIF_set_double, &send_secs, 0, 0 },
+	{ "e", "extensions", 0, "Show ICMP extensions (if present), "
+			    "including MPLS",
+			    CLIF_set_flag, &extension, 0, CLIF_ABBREV },
 	{ "A", "as-path-lookups", 0, "Perform AS path lookups in routing "
 			    "registries and print results directly after "
 			    "the corresponding addresses",
@@ -498,6 +511,12 @@ static CLIF_option option_list[] = {
 	{ "P", "protocol", "prot", "Use raw packet of protocol %s "
 			    "for tracerouting", 
 			    set_raw, 0, 0, CLIF_EXTRA },
+	{ 0, "mtu", 0, "Discover MTU along the path being traced. "
+			    "Implies `-F -N 1'",
+			    CLIF_set_flag, &mtudisc, 0, CLIF_EXTRA },
+	{ 0, "back", 0, "Guess the number of hops in the backward path "
+			    "and print if it differs",
+			    CLIF_set_flag, &backward, 0, CLIF_EXTRA },
 	CLIF_VERSION_OPTION (version_string),
 	CLIF_HELP_OPTION,
 	CLIF_END_OPTION
@@ -506,10 +525,10 @@ static CLIF_option option_list[] = {
 static CLIF_argument arg_list[] = {
         { "host", "The host to traceroute to",
 				set_host, 0, CLIF_STRICT },
-	{ "packetlen", "Specify an alternate probe packet length "
-			"(default is " _TEXT(DEF_PACKET_LEN) ")."
-			" Useless for TCP SYN",
-				CLIF_arg_uint, &packet_len, 0 },
+	{ "packetlen", "The full packet length (default is the length of "
+			"an IP header plus " _TEXT(DEF_DATA_LEN) "). Can be "
+			"ignored or increased to a minimal allowed value",
+				CLIF_arg_int, &packet_len, 0 },
 	CLIF_END_ARGUMENT
 };
 
@@ -543,8 +562,6 @@ int main (int argc, char *argv[]) {
 		ex_error ("max hops cannot be more than " _TEXT(MAX_HOPS));
 	if (!probes_per_hop || probes_per_hop > MAX_PROBES)
 		ex_error ("no more than " _TEXT(MAX_PROBES) " probes per hop");
-	if (!sim_probes || sim_probes > max_hops * probes_per_hop)
-		ex_error ("sim hops out of range");
 	if (wait_secs < 0)
 		ex_error ("bad wait seconds `%g' specified", wait_secs);
 	if (packet_len > MAX_PACKET_LEN)
@@ -569,7 +586,7 @@ int main (int argc, char *argv[]) {
 		sim_probes = 1;
 
 
-	/*  make sure we don't std{in|,out,err} to open sockets  */
+	/*  make sure we don't std{in,out,err} to open sockets  */
 	make_fd_used (0);
 	make_fd_used (1);
 	make_fd_used (2);
@@ -577,6 +594,24 @@ int main (int argc, char *argv[]) {
 
 	init_ip_options ();
 
+	header_len = (af == AF_INET ? sizeof (struct iphdr)
+				    : sizeof (struct ip6_hdr)) +
+			rtbuf_len + ops->header_len;
+
+	if (mtudisc) {
+	    dontfrag = 1;
+	    sim_probes = 1;
+	    packet_len = MAX_PACKET_LEN;
+	}
+
+	if (packet_len < 0) {
+	    if (DEF_DATA_LEN >= ops->header_len)
+		    data_len = DEF_DATA_LEN - ops->header_len;
+	} else {
+	    if (packet_len >= header_len)
+		    data_len = packet_len - header_len;
+	}
+
 
 	num_probes = max_hops * probes_per_hop;
 	probes = calloc (num_probes, sizeof (*probes));
@@ -589,7 +624,7 @@ int main (int argc, char *argv[]) {
 		    exit (2);
 	}
 
-	if (ops->init (&dst_addr, dst_port_seq, packet_len) < 0)
+	if (ops->init (&dst_addr, dst_port_seq, &data_len) < 0)
 		ex_error ("trace method's init failed");
 
 
@@ -605,17 +640,17 @@ static void print_header (void) {
 
 	/*  Note, without ending new-line!  */
 	printf ("traceroute to %s (%s), %u hops max, %u byte packets",
-			dst_name, addr2str (&dst_addr), max_hops, packet_len);
+				dst_name, addr2str (&dst_addr), max_hops,
+				header_len + data_len);
 	fflush (stdout);
 }
 
+
 static void print_addr (sockaddr_any *res) {
 	const char *str;
 
-	if (!res->sa.sa_family) {
-		printf (" *");
+	if (!res->sa.sa_family)
 		return;
-	}
 
 	str = addr2str (res);
 
@@ -637,29 +672,47 @@ static void print_addr (sockaddr_any *res) {
 		printf (" [%s]", get_as_path (str));
 }
 
+
 static void print_probe (probe *pb) {
-	unsigned int idx, np;
+	unsigned int idx = (pb - probes);
+	unsigned int ttl = idx / probes_per_hop + 1;
+	unsigned int np = idx % probes_per_hop;
 
-	idx = (pb - probes);
-	np = idx % probes_per_hop;
+	if (np == 0)
+		printf ("\n%2u ", ttl);
 
-	if (np == 0) {
-	    int ttl;
 
-	    ttl = idx / probes_per_hop + 1;
+	if (!pb->res.sa.sa_family)
+		printf (" *");
+	else {
+	    int prn = !np;	/*  print if the first...  */
 
-	    printf ("\n%2u ", ttl);
-	    print_addr (&pb->res);
+	    if (np) {	    /*  ...and if differs with previous   */
+		probe *p;
+
+		/*  skip expired   */
+		for (p = pb - 1; np && !p->res.sa.sa_family; p--, np--) ;
+
+		if (!np ||
+		    !equal_addr (&p->res, &pb->res) ||
+		    (extension && p->ext != pb->ext &&
+			!(p->ext && pb->ext && !strcmp (p->ext, pb->ext))) ||
+		    (backward && p->recv_ttl != pb->recv_ttl)
+		)  prn = 1;
+	    }
+
+	    if (prn) {
+		print_addr (&pb->res);
+
+		if (pb->ext)  printf (" <%s>", pb->ext);
+
+		if (backward && pb->recv_ttl) {
+		    int hops = ttl2hops (pb->recv_ttl);
+		    if (hops != ttl)  printf (" '-%d'", hops);
+		}
+	    }
 	}
-	else {	/*  print if differs with previous   */
-	    probe *p;
 
-	    /*  skip expired   */
-	    for (p = pb - 1; np && !p->res.sa.sa_family; p--, np--) ;
-
-	    if (!np || !equal_addr (&p->res, &pb->res))
-		    print_addr (&pb->res);
-	}
 
 	if (pb->recv_time) {
 	    double diff = pb->recv_time - pb->send_time;
@@ -670,11 +723,13 @@ static void print_probe (probe *pb) {
 	if (pb->err_str[0])
 		printf (" %s", pb->err_str);
 
+
 	fflush (stdout);
 
 	return;
 }
 
+
 static void print_end (void) {
 
 	printf ("\n");
@@ -763,8 +818,6 @@ static void check_expired (probe *pb) {
 		in the modern routers and computers.
 		The idea comes from tracepath(1) routine.
 	    */
-#define ttl2hops(X)	(((X) <= 64 ? 65 : ((X) <= 128 ? 129 : 256)) - (X))
-
 	    back_hops = ttl2hops (fp->recv_ttl);
 
 	    /*  It is possible that the back path differs from the forward
@@ -1025,7 +1078,7 @@ void tune_socket (int sk) {
 
 
 void parse_icmp_res (probe *pb, int type, int code, int info) {
-	char *str = "";
+	char *str = NULL;
 	char buf[sizeof (pb->err_str)];
 
 	if (af == AF_INET) {
@@ -1058,7 +1111,7 @@ void parse_icmp_res (probe *pb, int type, int code, int info) {
 
 		    case ICMP_UNREACH_PORT:
 			    /*  dest host is reached   */
-			    str = NULL;
+			    str = "";
 			    break;
 
 		    case ICMP_UNREACH_PROTOCOL:
@@ -1115,7 +1168,7 @@ void parse_icmp_res (probe *pb, int type, int code, int info) {
 
 		    case ICMP6_DST_UNREACH_NOPORT:
 			    /*  dest host is reached   */
-			    str = NULL;
+			    str = "";
 			    break;
 
 		    default:
@@ -1131,12 +1184,12 @@ void parse_icmp_res (probe *pb, int type, int code, int info) {
 	}
 
 
-	if (str && !*str) {
+	if (!str) {
 	    snprintf (buf, sizeof (buf), "!<%u-%u>", type, code);
 	    str = buf;
 	}
 
-	if (str) {
+	if (*str) {
 	    strncpy (pb->err_str, str, sizeof (pb->err_str));
 	    pb->err_str[sizeof (pb->err_str) - 1] = '\0';
 	}
@@ -1147,10 +1200,83 @@ void parse_icmp_res (probe *pb, int type, int code, int info) {
 }
 
 
-void parse_cmsg (probe *pb, struct msghdr *msg) {
-	struct cmsghdr *cm;
+void probe_done (probe *pb) {
 
-	for (cm = CMSG_FIRSTHDR (msg); cm; cm = CMSG_NXTHDR (msg, cm)) {
+	if (pb->sk) {
+	    del_poll (pb->sk);
+	    close (pb->sk);
+	    pb->sk = 0;
+	}
+
+	pb->seq = 0;
+
+	pb->done = 1;
+}
+
+
+void recv_reply (int sk, int err, check_reply_t check_reply) {
+	struct msghdr msg;
+	sockaddr_any from;
+	struct iovec iov;
+	int n;
+	probe *pb;
+	char buf[1280];		/*  min mtu for ipv6 ( >= 576 for ipv4)  */
+	char *bufp = buf;
+	char control[1024];
+	struct cmsghdr *cm;
+	struct sock_extended_err *ee = NULL;
+
+
+	memset (&msg, 0, sizeof (msg));
+	msg.msg_name = &from;
+	msg.msg_namelen = sizeof (from);
+	msg.msg_control = control;
+	msg.msg_controllen = sizeof (control);
+	iov.iov_base = buf;
+	iov.iov_len = sizeof (buf);
+	msg.msg_iov = &iov;
+	msg.msg_iovlen = 1;
+
+
+	n = recvmsg (sk, &msg, err ? MSG_ERRQUEUE : 0);
+	if (n < 0)  return;
+
+
+	/*  when not MSG_ERRQUEUE, AF_INET returns full ipv4 header
+	    on raw sockets...
+	*/
+
+	if (!err &&
+	    af == AF_INET &&
+	    /*  XXX: Assume that the presence of an extra header means
+		that it is not a raw socket...
+	    */
+	    ops->header_len == 0
+	) {
+	    struct iphdr *ip = (struct iphdr *) bufp;
+	    int hlen;
+
+	    if (n < sizeof (struct iphdr))  return;
+
+	    hlen = ip->ihl << 2;
+	    if (n < hlen)  return;
+
+	    bufp += hlen;
+	    n -= hlen;
+	}
+
+
+	pb = check_reply (sk, err, &from, bufp, n);
+	if (!pb)  return;
+
+
+	if (!err)
+	    memcpy (&pb->res, &from, sizeof (pb->res));
+
+
+	/*  Parse CMSG stuff   */
+
+	for (cm = CMSG_FIRSTHDR (&msg); cm; cm = CMSG_NXTHDR (&msg, cm)) {
 
 	    if (cm->cmsg_level == SOL_SOCKET) {
 
@@ -1165,15 +1291,11 @@ void parse_cmsg (probe *pb, struct msghdr *msg) {
 		if (cm->cmsg_type == IP_TTL)
 			pb->recv_ttl = *((int *) CMSG_DATA (cm));
 		else if (cm->cmsg_type == IP_RECVERR) {
-		    struct sock_extended_err *ee;
 
 		    ee = (struct sock_extended_err *) CMSG_DATA (cm);
 
-		    if (ee->ee_origin == SO_EE_ORIGIN_ICMP) {
-			memcpy (&pb->res, SO_EE_OFFENDER (ee), sizeof(pb->res));
-			parse_icmp_res (pb, ee->ee_type, ee->ee_code,
-								ee->ee_info);
-		    }
+		    if (ee->ee_origin != SO_EE_ORIGIN_ICMP)
+			    ee = NULL;
 		}
 	    }
 	    else if (cm->cmsg_level == SOL_IPV6) {
@@ -1181,24 +1303,67 @@ void parse_cmsg (probe *pb, struct msghdr *msg) {
 		if (cm->cmsg_type == IPV6_HOPLIMIT)
 			pb->recv_ttl = *((int *) CMSG_DATA (cm));
 		else if (cm->cmsg_type == IPV6_RECVERR) {
-		    struct sock_extended_err *ee;
 
 		    ee = (struct sock_extended_err *) CMSG_DATA (cm);
 
-		    if (ee->ee_origin == SO_EE_ORIGIN_ICMP6) {
-			memcpy (&pb->res, SO_EE_OFFENDER (ee), sizeof(pb->res));
-			parse_icmp_res (pb, ee->ee_type, ee->ee_code,
-								ee->ee_info);
-		    }
+		    if (ee->ee_origin != SO_EE_ORIGIN_ICMP6)
+			    ee = NULL;
 		}
 	    }
 	}
 
+	if (ee) {
+	    memcpy (&pb->res, SO_EE_OFFENDER (ee), sizeof(pb->res));
+	    parse_icmp_res (pb, ee->ee_type, ee->ee_code, ee->ee_info);
+	}
+
 
 	if (!pb->recv_time)
 	    pb->recv_time = get_time ();
 
-	return;
+
+	if (ee &&
+	    mtudisc &&
+	    ee->ee_info >= header_len &&
+	    ee->ee_info < header_len + data_len
+	) {
+	    data_len = ee->ee_info - header_len;
+
+	    probe_done (pb);
+
+	    /*  clear this probe (as actually the previous hop answers here)
+	      but fill its `err_str' by the info obtained. Ugly, but easy...
+	    */
+	    memset (pb, 0, sizeof (*pb));
+	    snprintf (pb->err_str, sizeof(pb->err_str)-1, "F=%d", ee->ee_info);
+
+	    return;
+	}
+
+
+	if (ee &&
+	    extension &&
+	    header_len + n >= (128 + 8) &&	/*  at least... (rfc4884)  */
+	    header_len <= 128 &&	/*  paranoia   */
+	    ((af == AF_INET && (ee->ee_type == ICMP_TIME_EXCEEDED ||
+				ee->ee_type == ICMP_DEST_UNREACH ||
+				ee->ee_type == ICMP_PARAMETERPROB)) ||
+	     (af == AF_INET6 && (ee->ee_type == ICMP6_TIME_EXCEEDED ||
+				 ee->ee_type == ICMP6_DST_UNREACH))
+	    )
+	) {
+	    int step;
+	    int offs = 128 - header_len;
+
+	    if (n > data_len)  step = 0;	/*  guaranteed at 128 ...  */
+	    else
+		step = af == AF_INET ? 4 : 8;
+
+	    handle_extensions (pb, bufp + offs, n - offs, step);
+	}
+
+
+	probe_done (pb);
 }
 
 
@@ -1298,9 +1463,13 @@ int do_send (int sk, const void *data, size_t len, const sockaddr_any *addr) {
 	else
 	    res = sendto (sk, data, len, 0, &addr->sa, sizeof (*addr));
 
-	if (res < 0 &&
-	    errno != ENOBUFS && errno != EAGAIN && errno != EMSGSIZE
-	)  error ("send");	/*  not recoverable   */
+	if (res < 0) {
+	    if (errno == ENOBUFS || errno == EAGAIN)
+		    return res;
+	    if (errno == EMSGSIZE)
+		    return 0;	/*  icmp will say more...  */
+	    error ("send");	/*  not recoverable   */
+	}
 
 	return res;
 }
